"""
This module computes API requests.
"""
import functools
import os
from datetime import datetime, timedelta
import time

import json
import requests

from multiprocessing import Pool
from navitia_client.utils import important_print
import navitia_client

_RETRIABLE_STATUSES = set([500, 503, 504])


class Client(object):
    """
    Performs requests to the navitia API web services.
    """

    def __init__(self, user, password="", retry_timeout=20, core_url='https://api.navitia.io/v1/', region=None):
        self.core_url = core_url
        self.user = user
        self.password = password
        self.retry_timeout = retry_timeout
        self.requested_urls = []
        if region:
            self.region = region

    def set_region(self, region):
        if not isinstance(region, str):
            raise ValueError("Region must be a string")
        self.region = region

    def _get(self, url, extra_params=None, verbose=False, first_request_time=None, retry_counter=0, ignore_fail=False):
        if verbose and not first_request_time:
            print("Import on url %s " % url)

        if not first_request_time:
            first_request_time = datetime.now()

        elapsed = datetime.now() - first_request_time
        if elapsed > timedelta(seconds=self.retry_timeout):
            raise navitia_client.exceptions.Timeout()

        if retry_counter > 0:
            # 0.5 * (1.5 ^ i) is an increased sleep time of 1.5x per iteration,
            # starting at 0.5s when retry_counter=0. The first retry will occur
            # at 1, so subtract that first.
            delay_seconds = 0.5 * 1.5 ** (retry_counter - 1)
            time.sleep(delay_seconds)

        full_url = os.path.join(self.core_url, url)

        try:
            response = requests.get(
                url=full_url, auth=(self.user, self.password), params=(extra_params or {}))
            self.requested_urls.append(response.url)

        except requests.exceptions.Timeout:
            if not ignore_fail:
                raise navitia_client.exceptions.Timeout()
            else:
                return False
        except Exception as e:
            if not ignore_fail:
                raise navitia_client.exceptions.TransportError(e)
            else:
                return False

        # Warn if not 200
        if response.status_code != 200:
            print("WARNING: response status_code is %s" % response.status_code)

        if response.status_code in _RETRIABLE_STATUSES:
            # Retry request.
            print("WARNING: retry number %d" % retry_counter)
            return self._get(url=url, extra_params=extra_params, first_request_time=first_request_time, retry_counter=retry_counter + 1, verbose=verbose, ignore_fail=ignore_fail)

        return response

    def _get_single_page_multiprocess(self, url, page, count, extra_params=None, verbose=False):
        """
        Return tuple: (page_number,response)
        Needed to perform multiprocessed requests.
        """
        if verbose:
            print("Import on page %d " % page)
        # Pagination parameters of request
        pagination_params = {
            "start_page": page,
            "count": count,
        }
        parameters = pagination_params
        try:
            for key, value in extra_params.items():
                parameters[key] = value
        except:
            # no extra param
            pass

        # Ignore fail: if one request fails, we still want others to be
        # computed
        # Verbose for first page only
        if page != 0:
            verbose = False
        response = self._get(
            url=url, extra_params=parameters, ignore_fail=True, verbose=verbose)

        return (page, response)

    def _extract_nbr_results(self, response):
        """
        Out of a request response, finds total number of results
        """
        parsed = json.loads(response.text)
        try:
            nbr_results = parsed["pagination"]["total_result"]
            return nbr_results
        except KeyError:
            # No pagination in page
            print("WARNING: not able to extract pagination out of first request.")
            return False

    def _get_multipage(self, url, page_limit=10, count=100, extra_params=None, verbose=False, log=False):
        """
        url: str
        #

        page_limit: int
        #

        count: int
        #

        extra_params: dict
        #

        verbose: boolean
        #

        log: boolean
        #

        ############
        return: dict
            key: page number
            value: response object
        """
        responses = {}

        # First page
        page, response = self._get_single_page_multiprocess(
            url=url, page=0, count=count, verbose=verbose, extra_params=extra_params)
        responses[page] = response

        # Find number of requests to make
        nbr_results = self._extract_nbr_results(responses[0])
        if not nbr_results:
            # If unable to get pagination count, returns first response
            return responses

        # Find number of pages to query
        blocs = nbr_results // count
        page_limit = min(page_limit, blocs + 1)
        if verbose:
            important_print("There are %d elements with %d elements per page. Limit is %d." % (
                nbr_results, count, page_limit))

        # Query other pages, multiprocessing
        pages = range(1, page_limit)
        pool = Pool(processes=30)
        n = len(pages)
        all_parameters = zip([self] * n, [url] * n,
                             pages, [count] * n, [extra_params or {}] * n, [verbose] * n)

        list_tuples = pool.map(unwrap_self, all_parameters)
        pool.close()
        pool.join()
        # Add results to initial results (first page)
        responses.update(dict(list_tuples).items())

        # Save logs
        if log:
            log = {}
            log["date"] = str(datetime.now().isoformat())
            log["url"] = url
            log["count_per_page"] = count
            log["page_limit"] = page_limit
            log["nbr_results"] = nbr_results
            responses["log"] = log
        if verbose:
            print("Multipage finished")
        return responses


def unwrap_self(arg, **kwarg):
    """ To unwrap self in multiprocessing mode
    http://www.rueckstiess.net/research/snippets/show/ca1d7d90
    """
    return Client._get_single_page_multiprocess(*arg, **kwarg)


def make_api_method(func):
    """
    Provides a single entry point for modifying all API methods.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result
    return wrapper

from navitia_client.journeys import journeys
from navitia_client.route_schedules import route_schedules
from navitia_client.raw import raw
from navitia_client.inverted_geocoding import inverted_geocoding
from navitia_client.explore import explore
from navitia_client.departures import departures

Client.journeys = make_api_method(journeys)
Client.route_schedules = make_api_method(route_schedules)
Client.raw = make_api_method(raw)
Client.inverted_geocoding = make_api_method(inverted_geocoding)
Client.explore = make_api_method(explore)
Client.departures = make_api_method(departures)




"""Tests for the object departures module."""

import responses
# initialize package, and does not mix up names
import test as _test
import navitia_client
import requests


class DeparturesTest(_test.TestCase):

    def setUp(self):
        self.user = 'leo'
        self.core_url = "https://api.navitia.io/v1/"
        self.client = navitia_client.Client(self.user)
        self.coords = '2.333333;48.866667'

    def test_no_region_nor_coords(self):
        # Should raise error if no region nor coords specified
        pass



"""
http://doc.navitia.io/#inverted-geocoding
Also known as /coords service.

Very simple service: you give Navitia some coordinates, it answers you

    your detailed postal address
    the right Navitia “coverage” which allows you to access to all known local mobility services


Accesses
url                                        Result
coords/{lon;lat}                           Detailed address point
coverage/{region_id}/coords/{lon;lat}      Detailed address point

You can also combine /coords with other filter as :

    get POIs near a coordinate
        https://api.navitia.io/v1/coverage/fr-idf/coords/2.377310;48.847002/pois?distance=1000
    get specific POIs near a coordinate
        https://api.navitia.io/v1/coverage/fr-idf/poi_types/poi_type:amenity:bicycle_rental/coords/2.377310;48.847002/pois?distance=1000

"""
import os


def inverted_geocoding(client, coords=None, lat=None, lon=None, pois=None, distance=None, region=None, verbose=False):

    # By default do not use region: coords/{lon;lat}
    if not region:
        base_url = "coords"

    # Else: coverage/{region_id}/coords/{lon;lat}
    else:
        if region == True:
            if not hasattr(client, 'region'):
                raise ValueError(
                    "You must specifiy region, either here or in client")
            else:
                used_region = client.region

        elif isinstance(region, str):
            # maybe check
            used_region = region

        base_url = os.path.join("coverage", used_region, "coords")

    extra_params = {}

    # coords, or lat and lon
    arg_count = sum(1 for x in [coords, lat and lon] if x)
    if arg_count != 1:
        raise ValueError(
            "Needs coords, or latitude and longitude.")

    if coords:
        # TODO check if coords ok
        url = os.path.join(base_url, coords)

    if lat and lon:
        # TODO check if lat and lon ok
        url = os.path.join(base_url, str(lon) + ";" + str(lat))

    if pois:
        # TODO check if pois ok
        url = os.path.join(url, "pois")

    if distance:
        # TODO check distance
        extra_params["distance"] = distance
    return client._get(url=url, extra_params=extra_params, verbose=verbose)




"""Tests for the object exploration module: explore."""

import responses
# initialize package, and does not mix up names
import test as _test
import navitia_client
import requests


class ExploreTest(_test.TestCase):

    def setUp(self):
        self.user = 'leo'
        self.core_url = "https://api.navitia.io/v1/"
        self.client = navitia_client.Client(self.user)
        self.coords = '2.333333;48.866667'

    def test_non_authorized_ressource(self):
        client = self.client
        client.set_region("sncf")
        false_ressource = 'doesnotexist'
        with self.assertRaises(ValueError):
            client.explore(false_ressource)

    def test_no_region_nor_coords(self):
        # Should raise error if no region nor coords specified
        client = self.client
        ressource = 'disruptions'
        with self.assertRaises(ValueError):
            client.explore(ressource)

    @responses.activate
    def test_explore_ressource_region(self):
        client = self.client
        client.set_region("sncf")
        url = 'https://api.navitia.io/v1/coverage/sncf/disruptions'
        responses.add(responses.GET, url,
                      body='{"disruptions": [], "links": [], "pagination": []}',
                      status=200, content_type='application/json')

        self.client.explore("disruptions")
        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(url, responses.calls[0].request.url)

    @responses.activate
    def test_explore_ressource_coords(self):
        client = self.client
        client.set_region("sncf")
        url = 'https://api.navitia.io/v1/coverage/2.333333;48.866667/lines'
        responses.add(responses.GET, url,
                      body='{"disruptions": [], "links": [], "pagination": []}',
                      status=200, content_type='application/json')

        self.client.explore("lines", coords=self.coords, distance=1000)
        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(url + "?distance=1000",
                            responses.calls[0].request.url)




"""Tests for the inverted-geocoding module."""

import responses
# initialize test package __init__, and does not mix up names
import test as _test
import navitia_client
import requests


class InvertedGeocodingTest(_test.TestCase):

    def setUp(self):
        self.user = 'leo'
        self.core_url = "https://api.navitia.io/v1/"
        self.client = navitia_client.Client(self.user)
        self.lat = '48.866667'
        self.lon = '2.333333'
        self.coords = '2.333333;48.866667'

    def test_latlon_and_coords(self):
        # Should fail, need one or the other
        client = self.client
        with self.assertRaises(ValueError):
            client.inverted_geocoding(
                lat=self.lat, lon=self.lon, coords=self.coords)
        # another try with region given, should fail as well
        client.set_region("sncf")
        with self.assertRaises(ValueError):
            client.inverted_geocoding(
                lat=self.lat, lon=self.lon, coords=self.coords)

    @responses.activate
    def test_coords_request(self):
        # Without region: should work: coords/{lon;lat}
        client = self.client
        url1 = "https://api.navitia.io/v1/coords/2.333333;48.866667"
        responses.add(responses.GET, url1,
                      body='yo',
                      status=200, content_type='application/json')
        client.inverted_geocoding(coords=self.coords)
        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(url1, responses.calls[0].request.url)

        # With region: should work: coverage/{region_id}/coords/{lon;lat}
        client.set_region("sncf")
        url2 = "https://api.navitia.io/v1/coverage/sncf/coords/2.333333;48.866667"
        responses.add(responses.GET, url2,
                      body='yo',
                      status=200, content_type='application/json')
        client.inverted_geocoding(coords=self.coords, region=True)
        self.assertEqual(2, len(responses.calls))
        self.assertURLEqual(url2, responses.calls[1].request.url)

    @responses.activate
    def test_lat_lon_request(self):
        # Without region: should work: coords/{lon;lat}
        client = self.client
        url1 = "https://api.navitia.io/v1/coords/2.333333;48.866667"
        responses.add(responses.GET, url1,
                      body='yo',
                      status=200, content_type='application/json')
        client.inverted_geocoding(lat=self.lat, lon=self.lon)
        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(url1, responses.calls[0].request.url)

        # With region: should work: coverage/{region_id}/coords/{lon;lat}
        client.set_region("sncf")
        url2 = "https://api.navitia.io/v1/coverage/sncf/coords/2.333333;48.866667"
        responses.add(responses.GET, url2,
                      body='yo',
                      status=200, content_type='application/json')
        client.inverted_geocoding(lat=self.lat, lon=self.lon, region=True)
        self.assertEqual(2, len(responses.calls))
        self.assertURLEqual(url2, responses.calls[1].request.url)



"""
Departures
http://doc.navitia.io/#departures

Also known as /departures service.

This endpoint retrieves a list of departures from a specific datetime of a selected object. Departures are ordered chronologically in ascending order as:


url                                                 Result
/coverage/{region_id}/{resource_path}/departures    List of the next departures, multi-route oriented, only time sorted (no grouped by stop_point/route here)

/coverage/{lon;lat}/coords/{lon;lat}/departures     List of the next departures, multi-route oriented, only time sorted (no grouped by stop_point/route here)
"""
import os


def departures(client, collection_name=None, object_id=None, coords=None, region=None, extra_params=None, verbose=False):
    # Construct url
    if coords and region:
        raise ValueError(
            "Cannot specifiy both coords and region, you must choose one.")
    if coords:
        # TODO: check coords format
        # /coverage/{lon;lat}/coords/{lon;lat}/departures
        url = os.path.join("coverage", coords, "coords",
                           coords, "departures")
    else:
        # /coverage/{region_id}/{resource_path}/departures
        # First choose region
        if not region and not hasattr(client, 'region'):
            raise ValueError(
                "You must specifiy coords or region, either here or in client")
        elif region:
            if isinstance(region, str):
                # region argument overrides client specified region
                used_region = region
            else:
                raise ValueError("Region must be a string")
        elif not region and hasattr(client, 'region'):
            # Takes already specified region
            used_region = client.region
        else:
            # shouldn't be possible
            raise ValueError("Weird error, caused by region")
        # /coverage/{region_id}/{collection_name}
        if not object_id or not collection_name:
            raise ValueError("of correct type")
        url = os.path.join("coverage", used_region,
                           collection_name, object_id, "departures")

    return client._get(url=url, extra_params=extra_params, verbose=verbose)



"""
Public Transportation Objects exploration
http://doc.navitia.io/#public-transportation-objects-exploration

Also known as /networks, /lines, /stop_areas… services.

Once you have selected a region, you can explore the public transportation objects easily with these APIs. You just need to add at the end of your URL a collection name to see every objects within a particular collection. To see an object detail, add the id of this object at the end of the collection’s URL. The paging arguments may be used to paginate results.
Accesses
url                                                 Result
/coverage/{region_id}/{collection_name}             Collection of objects of a region
/coverage/{region_id}/{collection_name}/{object_id} Information about a specific region
/coverage/{lon;lat}/{collection_name}               Collection of objects of a region
/coverage/{lon;lat}/{collection_name}/{object_id}   Information about a specific region

Focus on first two usages (coord service doesn't work for sncf)

Either coords, or region

"""
import os


def explore(client, collection_name, coords=None, region=None, depth=None, distance=None, disable_geojson=None, extra_params=None, specific_object=None, multipage=None, count_per_page=None, page_limit=None, url_suffix=None, verbose=False):

    if specific_object and multipage:
        raise UserWarning("Specific objects return usually one page.")

    # Check accepted_collections
    accepted_collections = {
        "networks",
        "lines",
        "routes",
        "stop_points",
        "stop_areas",
        "commercial_modes",
        "physical_modes",
        "companies",
        "vehicle_journeys",
        "disruptions"
    }
    if collection_name not in accepted_collections:
        raise ValueError("Collection name provided is not supported.")

    # Construct url
    if coords and region:
        raise ValueError(
            "Cannot specifiy both coords and region, you must choose one.")
    if coords:
        # TODO: check coords format
        # /coverage/{lon;lat}/{collection_name}
        url = os.path.join("coverage", coords, collection_name)
    else:
        # First choose region
        if not region and not hasattr(client, 'region'):
            raise ValueError(
                "You must specifiy coords or region, either here or in client")
        elif region:
            if isinstance(region, str):
                # region argument overrides client specified region
                used_region = region
            else:
                raise ValueError("Region must be a string")
        elif not region and hasattr(client, 'region'):
            # Takes already specified region
            used_region = client.region
        else:
            # shouldn't be possible
            raise ValueError("Weird error, caused by region")
        # /coverage/{region_id}/{collection_name}
        url = os.path.join("coverage", used_region, collection_name)

    # Specific object
    if specific_object:
        url = os.path.join(url, specific_object)

    # Url suffix
    if url_suffix:
        url = os.path.join(url, url_suffix)
    # FILTERS
    # OK: depth, distance, disable_geojson
    # Not taken into account because too specific: odt level, headsign, since
    # / until
    # They can still be taken into account in extra_params argument
    params = {}

    if depth:
        try:
            depth = int(depth)
            params["depth"] = depth
        except ValueError:
            raise ValueError("Depth must be int")

    if distance:
        if not coords:
            raise ValueError("Distance works only with coords")
        try:
            distance = int(distance)
            params["distance"] = distance
        except ValueError:
            raise ValueError("Distance must be int")

    if disable_geojson:
        params["disable_geojson"] = True

    all_params = {}

    try:
        for key, value in params.items():
            all_params[key] = value
    except:
        pass

    try:
        for key, value in extra_params.items():
            all_params[key] = value
    except:
        pass

    if not multipage:
        return client._get(url=url, extra_params=all_params, verbose=verbose)

    else:
        return client._get_multipage(url=url, page_limit=page_limit, count=count_per_page, extra_params=all_params, verbose=verbose)



"""
Tests for client module.

Uses responses package to mock requests package.
https://pypi.python.org/pypi/responses

"""

import responses
import time

import navitia_client
import test as _test
import requests


class ClientTest(_test.TestCase):

    def setUp(self):
        self.user = 'leo'
        self.core_url = "https://api.navitia.io/v1/"
        self.client = navitia_client.Client(self.user)

    @responses.activate
    def test_simple_request(self):
        responses.add(responses.GET,
                      "https://api.navitia.io/v1/coverage/sncf/disruptions",
                      body='yeah',
                      status=200,
                      content_type="application/json")

        self.client.raw("coverage/sncf/disruptions")
        self.assertEqual(1, len(responses.calls))

    # could test retriable status, and parameters




"""Tests for the route_schedules module."""

import responses
# initialize test package __init__, and does not mix up names
import test as _test
import navitia_client
import requests


class RouteSchedulesTest(_test.TestCase):

    def setUp(self):
        self.user = 'leo'
        self.core_url = "https://api.navitia.io/v1/"
        self.client = navitia_client.Client(self.user)
        self.coords = '2.333333;48.866667'
        self.datetime = '20161221T000000'

    def test_no_region(self):
        # Should fail, needs one region
        client = self.client
        with self.assertRaises(ValueError):
            client.route_schedules(raw="anything")

    def test_wrong_datafreshness(self):
        # Should fail, only base_schedule or realtime
        client = self.client
        client.set_region("sncf")
        with self.assertRaises(ValueError):
            client.route_schedules(raw="anything", data_freshness="lalala")

    def test_multiple_ressources_arguments(self):
        # Should fail, only one accepted out of line, raw, route etc..
        client = self.client
        client.set_region("sncf")
        with self.assertRaises(ValueError):
            client.route_schedules(raw="anything", line="line")




"""Tests for the journeys module."""

import responses
# initialize test package __init__, and does not mix up names
import test as _test
import navitia_client
import requests


class JourneysTest(_test.TestCase):

    def setUp(self):
        self.user = 'leo'
        self.core_url = "https://api.navitia.io/v1/"
        self.client = navitia_client.Client(self.user)
        self.lat = '48.866667'
        self.lon = '2.333333'
        self.coords = '2.333333;48.866667'
        self.datetime = '20161221T000000'

    def test_no_region(self):
        # Should fail, needs one region
        client = self.client
        with self.assertRaises(ValueError):
            client.journeys(origin=self.coords)

    def test_wrong_datetime_represents(self):
        # Should fail, only departure or arrival
        client = self.client
        client.set_region("sncf")
        with self.assertRaises(ValueError):
            client.journeys(
                origin=self.coords, datetime=self.datetime, datetime_represents="lalala")

    def test_datetime_represents_without_datetime(self):
        # Should fail, datetime_represents needs datetime
        client = self.client
        client.set_region("sncf")
        with self.assertRaises(ValueError):
            client.journeys(origin=self.coords, datetime_represents="lalala")

    def test_wrong_datafreshness(self):
        # Should fail, only base_schedule or realtime
        client = self.client
        client.set_region("sncf")
        with self.assertRaises(ValueError):
            client.journeys(origin=self.coords, data_freshness="lalala")
